/* s3emul - "secu3 emulator"
   Copyright (C) 2016 andreika. Ukraine, Kiev

   The following code is a derivative work of the code from the SECU-3 project, 
   which is licensed GPLv3. This code therefore is also licensed under the terms 
   of the GNU Public License, verison 3.

   PORTIONS OF THIS CODE TAKEN FROM:

   SECU-3  - An open source, free engine control unit
   Copyright (C) 2007 Alexey A. Shabelnikov. Ukraine, Kiev

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   contacts:
              http://secu-3.org
              email: shabelnikov@secu-3.org
*/

#include <iostream>
#include <fstream>
#include <string>
#include <regex>
#include <array>

#include "s3emul.h"
#include "port.h"


#define DUMP_EEPROM
//#define PRINT_DEBUG_S3

static struct ecudata_t d;
static struct fw_data_t fw;
static f_data_t realtime_table;
static int curTblIdx = -1;

static uint8_t *dPtr = (uint8_t *)&d;

static std::vector<CSVRecord> logRecords;

static s3emul_params p;
static uartstate_t uart;

static std::string paramsFile;


struct
{
	int mode;
	const char *modeName;
} sendModes[] = 
{
	{ CHANGEMODE, "change mode" },
	{ BOOTLOADER, "boot loader" },
	{ TEMPER_PAR, "temperature" },
	{ CARBUR_PAR, "carb." },
	{ IDLREG_PAR, "idling reg." },
	{ ANGLES_PAR, "advance angle" },
	{ FUNSET_PAR, "functions" },
	{ STARTR_PAR, "engine start" },
	{ FNNAME_DAT, "names of func." },
	{ SENSOR_DAT, "sensors data" },
	{ ADCCOR_PAR, "ADC corr." },
	{ ADCRAW_DAT, "raw values" },
	{ CKPS_PAR, "CKP sensor" },
	{ OP_COMP_NC, "suspended op" },
	{ CE_ERR_CODES, "CE codes" },
	{ KNOCK_PAR, "knock chip" },
	{ CE_SAVED_ERR, "CE EEPROM codes" },
	{ FWINFO_DAT, "firmware info" },
	{ MISCEL_PAR, "misc." },
	{ EDITAB_PAR, "RT tables" },
	{ ATTTAB_PAR, "attenuator map" },
	{ RPMGRD_PAR, "RPM grid" },
	{ DBGVAR_DAT, "debug" },
	{ DIAGINP_DAT, "diag. input" },
	{ DIAGOUT_DAT, "diag. output" },
	{ CHOKE_PAR, "choke" },
	{ SECUR_PAR, "security" },
	{ UNIOUT_PAR, "uni.outputs" },
	{ INJCTR_PAR, "injection" },
	{ LAMBDA_PAR, "lambda" },
	{ ACCEL_PAR, "accel.enrichment" },
	{ GASDOSE_PAR, "gas dose" },
	{ SIGINF_DAT, "signature" },
	{ -1, "" }
};

static std::string sendModeName[256];

void initRTTables()
{
	log("* Init params...\n");

	// copy from firmware/eeprom
	memcpy(&d.param, &fw.def_param, sizeof(params_t));

	if (curTblIdx < TABLES_NUMBER_PGM)
	{
		memcpy(&d.tables_ram, &fw.tables[curTblIdx], sizeof(f_data_t)); // from firmware
	}
	else
	{
		memcpy(&d.tables_ram, &realtime_table, sizeof(f_data_t));   // from EEPROM
	}
}

void updateFromLogRecord(double *fields)
{
	const SensorParam *sensorParams = getSensorParams();

	for (int i = 0; sensorParams[i].name != nullptr; i++)
	{
		if (sensorParams[i].logIdx < 0)
			continue;
		assert(sensorParams[i].logIdx < maxNumFields);
		double f = fields[sensorParams[i].logIdx];
		int res = round_int(f * sensorParams[i].s3.coef + sensorParams[i].s3.add);

		switch (sensorParams[i].s3.size)
		{
		case -1: *((int8_t *)&dPtr[sensorParams[i].s3.offs]) = (int8_t)res; break;
		case -2: *((int16_t *)&dPtr[sensorParams[i].s3.offs]) = (int16_t)res; break;
		case -4: *((int32_t *)&dPtr[sensorParams[i].s3.offs]) = (int32_t)res; break;
		case 1: *((uint8_t *)&dPtr[sensorParams[i].s3.offs]) = (uint8_t)res; break;
		case 2: *((uint16_t *)&dPtr[sensorParams[i].s3.offs]) = (uint16_t)res; break;
		case 4: *((uint32_t *)&dPtr[sensorParams[i].s3.offs]) = (uint32_t)res; break;
		}
	}

	int tI = getTableIndex(d, curTblIdx);
	if (tI != curTblIdx)
	{
		curTblIdx = tI;
		initRTTables();
	}
}

int findFwOption(char *opt)
{
	return (strstr(p.options.c_str(), opt) != nullptr) ? 1 : 0;
}

bool initS3Params()
{
	memset(&d, 0, sizeof(d));
	memset(&fw, 0, sizeof(fw));

	int i;
	for (i = 0; i < 256; i++)
		sendModeName[i] = std::string();
	for (i = 0; sendModes[i].mode >= 0; i++)
	{
		sendModeName[sendModes[i].mode] = std::string(sendModes[i].modeName);
	}

	curTblIdx = -1;

	if (p.logFileName[0] == '\0')
	{
		log("Setting parameters from config file...\n");
	}

	strcpy((char *)fw.fw_signature_info, "S3EMUL " S3EMUL_VERSION);
	fw.cddata.config = (findFwOption("DWELL_CONTROL") << 7) | (findFwOption("COOLINGFAN_PWM") << 8) | (findFwOption("REALTIME_TABLES") << 9) | (findFwOption("ICCAVR_COMPILER") << 10) |
		(findFwOption("AVRGCC_COMPILER") << 11) | (findFwOption("DEBUG_VARIABLES") << 12) | (findFwOption("PHASE_SENSOR") << 13) | (findFwOption("PHASED_IGNITION") << 14) |
		(findFwOption("FUEL_PUMP") << 15) | (findFwOption("THERMISTOR_CS") << 16) | (1 << 17) | (findFwOption("DIAGNOSTICS") << 18) | (findFwOption("HALL_OUTPUT") << 19) |
		(findFwOption("REV9_BOARD") << 20) | (findFwOption("STROBOSCOPE") << 21) | (findFwOption("SM_CONTROL") << 22) | (findFwOption("VREF_5V") << 23) |
		(findFwOption("HALL_SYNC") << 24) | (findFwOption("UART_BINARY") << 25) | (findFwOption("CKPS_2CHIGN") << 26) | (1 << 27) |
		(findFwOption("FUEL_INJECT") << 28) | (findFwOption("GD_CONTROL") << 29) | (findFwOption("CARB_AFR") << 30) | (findFwOption("CKPS_NPLUS1") << 31);

	static const int16_t rpg_grid[] = { 600, 720, 840, 990, 1170, 1380, 1650, 1950, 2310, 2730, 3210, 3840, 4530, 5370, 6360, 7500 };
	memcpy(fw.exdata.rpm_grid_points, rpg_grid, sizeof(rpg_grid));

	const SensorParam *sensorParams = getSensorParams();
	std::string cfgName = openIni(paramsFile);

	for (int i = 0; sensorParams[i].name != nullptr; i++)
	{
		if (sensorParams[i].name[0] == '\0')    // empty - not used for config
			continue;
		std::string buf = getIniString("Sensors", sensorParams[i].name, "", cfgName);
		double f = atof(buf.c_str());
		int res = round_int(f * sensorParams[i].s3.coef + sensorParams[i].s3.add);

		switch (sensorParams[i].s3.size)
		{
		case -1: *((int8_t *)&dPtr[sensorParams[i].s3.offs]) = (int8_t)res; break;
		case -2: *((int16_t *)&dPtr[sensorParams[i].s3.offs]) = (int16_t)res; break;
		case -4: *((int32_t *)&dPtr[sensorParams[i].s3.offs]) = (int32_t)res; break;
		case 1: *((uint8_t *)&dPtr[sensorParams[i].s3.offs]) = (uint8_t)res; break;
		case 2: *((uint16_t *)&dPtr[sensorParams[i].s3.offs]) = (uint16_t)res; break;
		case 4: *((uint32_t *)&dPtr[sensorParams[i].s3.offs]) = (uint32_t)res; break;
		}
	}

	return true;
}

void sendToSoft()
{
	writePort(BoardOut, uart.send_buf, uart.send_size);
}

int emulS3(const s3emul_params & params)
{
	log("Starting SECU Emulator!\n");

	p = params;

	log("Opening BOARD serial port %s @ %d...\n", p.portOut[Secu].c_str(), p.baudOut[Secu]);
	if (!initPort(BoardOut, p.portOut[Secu], p.baudOut[Secu]))
	{
		log("ERROR!\n");
		return -4;
	}

	log("Emulating SECU-3 board...\n");

	if (!initS3Params())
	{
		return -2;
	}

	if (p.fwFileName[0] != '\0')
	{
		log("Reading SECU-3 firmware file...\n");
		if (!readFW(p.fwFileName, &fw, &realtime_table, &d.ecuerrors_saved_transfer))
		{
			log("ERROR!\n");
			return false;
		}

		// copy from firmware/eeprom
		memcpy(&d.param, &fw.def_param, sizeof(params_t));
	}

	if (p.eepromFileName[0] != '\0')
	{
		log("Reading SECU-3 EEPROM file...\n");
		if (!readFW(p.eepromFileName, &fw, &realtime_table, &d.ecuerrors_saved_transfer))
		{
			log("ERROR!\n");
			return false;
		}
		//d->ecuerrors_for_transfer = d.ecuerrors_saved_transfer;
	}

	if (p.logFileName[0] != '\0')
	{
		log("Reading log file %s...\n", p.logFileName.c_str());
		double duration;
		logRecords.clear();
		if (!readCSVFile(p.logFileName, logRecords, &duration))
		{
			log("ERROR reading log file!\n");
			return false;
		}

		log("Found %d records. Total duration: %d secs\n", logRecords.size(), round_int(duration));
	} else
	{
		curTblIdx = getTableIndex(d, curTblIdx);
		initRTTables();
	}

	log("Sending/reading packages...\n");

	init_s3_uart(uart, true);

	int pType = 0, sentPType = 0;

	size_t cnt = 0;
	double startTime = getTimeInSeconds();
	double printTime = 0.0;
	int printCnt = 0;

	for (;;)
	{
		if (getExiting())
			break;

		if (logRecords.size() > 0)
		{
			double curTime = getTimeInSeconds() - startTime;
			for (; cnt < logRecords.size() && logRecords[cnt].time < curTime; cnt++)
			{
				updateFromLogRecord(logRecords[cnt].fields);
				send_s3_client_packet(uart, &d, fw, pType);
				sendToSoft();
			}

			if (curTime - printTime > 2.0 || cnt >= logRecords.size())  // every 2 secs
			{
				log("  + %d records added (%d%%)\n", (cnt - printCnt), (cnt * 100 / logRecords.size()));
				printTime = curTime;
				printCnt = cnt;
			}
			if (cnt >= logRecords.size())
				break;
		}
		else
		{
			send_s3_client_packet(uart, &d, fw, pType);
			sendToSoft();
		}
		sentPType = pType;
		pType = 0;
		delay(1);

		// read some data
		int state = 0;
		for (int maxRead = 0; maxRead < UART_RECV_BUFF_SIZE; maxRead++)
		{
			int recv_size;
			uint8_t chr;
			if (!peekPortByte(BoardOut, &chr, &recv_size) || recv_size < 1)
				break;
			switch (state)
			{
			case 0:
				if (chr == '!') // start of a packet
				{
					state = 1;
					uart.recv_index = 0;
				}
				break;
			case 1:
				if (chr == '\r')
				{
					uart.recv_size = uart.recv_index;
					int descr = receive_s3_packet(uart, &d, fw);
					switch (descr)
					{
					case BOOTLOADER:
						log("* WARNING! BOOTLOADER mode not supported!\n");
						break;
					case OP_COMP_NC:
					{
#ifdef PRINT_DEBUG_S3
						log("OP_COMP_NC opcode=%d\n", _AB(d.op_actn_code, 0));
#endif
						int opCode = _AB(d.op_actn_code, 0);
						switch (opCode)
						{
						case OPCODE_READ_FW_SIG_INFO:
							if (sentPType != FWINFO_DAT)
								log("* Sending \"FW Info\" by request...\n");
							pType = FWINFO_DAT;
							break;
						case OPCODE_EEPROM_PARAM_SAVE:
						case OPCODE_SAVE_TABLSET:
						
							log("* EEPROM save request...\n");
#ifdef DUMP_EEPROM
							log("  Dumping EEPROM to %s\n", p.eepromSaveFileName.c_str());
							dumpEEPROM(p.eepromSaveFileName, &d);
#endif
							break;
						case OPCODE_DIAGNOST_ENTER:
							log("* WARNING! DIAGNOSTIC mode not supported!\n");
							break;
						}
						_AB(d.op_actn_code, 0) = 0; // processed ;-)
						break;
					}
					case EDITAB_PAR:
					{
						log("* Table %d changed!\n", get_s3_tblIdx(uart));
					}
					break;
					case CHANGEMODE:
					{
						static int printedSendMode = -1;
						if (uart.send_mode != printedSendMode)
						{
							log("* Sending \"%s\" by request...\n", sendModeName[uart.send_mode].c_str());
							printedSendMode = uart.send_mode;
						}
						send_s3_client_packet(uart, &d, fw, 0);
						sendToSoft();
						pType = 0;
					}
					break;
					default:
#ifdef PRINT_DEBUG_S3
						log("--%d ('%c') mode='%c'\n", descr, descr, uart.send_mode);
#endif
						;
					}
					maxRead = UART_RECV_BUFF_SIZE;
					break;
				}
				else
				{
					uart.recv_buf[uart.recv_index++] = chr;
				}
				break;
			}
		}
	}

	deInitPort(BoardOut);

	return 0;
}

